package org.codecop.basic

import scala.collection.mutable.{ArrayStack, Map}
/**
 * BASIC runtime memory for execution.
 * @author Peter Kofler
 */
trait BasicMemory {

   private val WAIT_FOR_BASIC_SPEED = 10 // wait 10 MS after each command
  
   // stack of program code
   private class Command(val lineNumber:Int, code: =>Unit) {
      def execute = code
      override def toString = lineNumber + " " + super.toString
   }
   private var programCode = List[Command]()

   // runtime data
   private val dVariables = Map[String, Double]()
   private val sVariables = Map[String, String]()
   private var jumpedLine = 0
   private var codeCounter = 0
   private var wasJump = false
   private class Loop(val codeLine:Int, val name:String, incTest: =>Boolean) {
     def next = incTest
   }
   private val loops = new ArrayStack[Loop]()

   protected val memory:LinesOfCode = new LinesOfCode {

      def addLine(line:Int)(code: =>Unit) {
         programCode += new Command(line, code)
         // could allow overwriting of numbers. 
         // when nuber is in between - insert
         // when number exists - delete old (and all -1) and insert
         // Problem: update all loop counters?
      }
      def addLoop(name:String)(incCheck: =>Boolean) {
        // codeCounter will point to the next command after loop begin
        val loopBegin = programCode.size + 1
        addLine(-1) {
           loops.push(new Loop(loopBegin, name, incCheck))
        }
      }
      
      def next(name:String) {
        if (loops.isEmpty) {
          throw new NextWithoutForError(name)
        }
        val loop = loops.pop
        if (name != null && name != loop.name) {
          throw new NextWithoutForError(name + " != " + loop.name)
        }
        if (loop.next) {
          // save again
          loops.push(loop) 
          codeCounter = loop.codeLine - 1 // +1 added by runtime after execution
        }
      }
      def jumpTo(line:Int) {
         jumpedLine = line
         wasJump = true
      }
      def end() {
         // jump to the end
         codeCounter = programCode.size
      }
      
      def set(name:String, value:Double) = dVariables.put(name, value)
      def get(name:String) = dVariables.get(name).get
      def setString(name:String, value:String) {
         if (!name.endsWith("$")) throw new IllegalArgumentException("wrong name " + name + " for String variable")
         sVariables.put(name, value)
      }
      def getString(name:String) = sVariables.get(name).get
   }
  
   // ---------- immediately commands ----------
   private def ready {
      println
      println("READY.")
   }
   def LIST {
     programCode.foreach(println)
     ready
   }

   def NEW {
      programCode = List[Command]()
      // ready
   }
  
   def RUN {
      dVariables.clear
      sVariables.clear
      codeCounter = 0
      wasJump = false
      loops.clear

      while (codeCounter < programCode.size || wasJump) {
         if (wasJump) {
            wasJump = false
            // find line with line number
            val trg = programCode.indices.find( programCode(_).lineNumber == jumpedLine )
            if (!trg.isDefined) {
               throw new LineNotFoundError(jumpedLine)
            }
            codeCounter = trg.get
         }
         
         programCode(codeCounter).execute
         codeCounter += 1

         if (WAIT_FOR_BASIC_SPEED>0) {
            Thread.sleep(WAIT_FOR_BASIC_SPEED)
         }
     }
     ready
  }
   
}
