package org.codecop.basic

/**
 * Parsing BASIC statements as inner DSL.
 * @author Peter Kofler
 */
trait BasicParser extends VariableEvaluation with BasicFunctions {

   private val LOG = false
   
   /** Interaction with the program space. Used by <code>VariableEvaluation</code>. */
   override protected def memory:LinesOfCode

   class NewStatement(line:Int) extends VariableAssignment[NewStatement] {
      private def this() = this(-1)
      override protected def memory = BasicParser.this.memory
      override protected def addToMemory(code: =>Unit) = {
         memory.addLine(line)(code)
         new NewStatement() // no line number
      }
      
      // ---------- simple commands ----------
      
      // END (of program)
      def END:Unit = addToMemory {
         if (LOG) println("exec. END")
         memory.end()
      }
      def GET(s:VariableStringExpression) = addToMemory {
         if (LOG) println("exec. GET " + s.fromVariable)
         // TODO this is not correct, would need KeyEvent or something like that
         val c = Console.readChar.toString
         memory.setString(s.fromVariable, if (c==null) "" else c)
      }
      // GOTO literal
      def GOTO(target:JumpTarget):Unit = addToMemory {
         if (LOG) println("exec. GOTO " + target.line)
         memory.jumpTo(target.line)
      }
      // INPUT string and number
      def INPUT(s:VariableStringExpression) = addToMemory {
         if (LOG) println("exec. INPUT " + s.fromVariable)
         memory.setString(s.fromVariable, Console.readLine.trim)
      }
      def INPUT(d:VariableDoubleExpression) = addToMemory {
         if (LOG) println("exec. INPUT " + d.fromVariable)
         memory.set(d.fromVariable, Console.readLine.trim.toDouble)
      }
      def PRINT:Unit = addToMemory { 
         println
      }
      def PRINT(msg:StringExpression) = addToMemory {
         if (LOG) println("exec. PRINT " + msg.value)
         println(msg.value)
         // allow following commands after PRINT
      }
      def ?(msg:StringExpression) = PRINT(msg)

      // ---------- extended commands ----------
      def IF(expr: => Boolean) = {
         // after the IF return an instance only allowing following THEN
         class ConsumeThen(expr: => Boolean) {
            def THEN(target:JumpTarget):Unit = addToMemory {
               if (LOG) println("exec. IF .. THEN " + target.line)
               if (expr) {
                  memory.jumpTo(target.line)
               }
            }
            // TODO THEN for assignments
         }
         new ConsumeThen(expr)
      }
      def FOR = {
         // after the FOR allow a variable assignment and ranges to be stored
         new Object() {
           def update(variable:VariableDoubleExpression, range:LoopRangeExpression) = {
              val usedVariable = variable.fromVariable
              val step = range.step
              addToMemory {
                 if (LOG) println("exec. FOR start: " + usedVariable + " = " + range.lower.value)
                 memory.set(usedVariable, range.lower.value)
              }
              memory.addLoop(usedVariable) {
                 // increment and check condition
                 val s = step.value
                 val counter = memory.get(usedVariable) + s
                 if ((s>0 && counter <= range.upper.value) || (s<0 && counter >= range.upper.value)) {
                    if (LOG) println("exec. FOR inc: " + counter)
                    memory.set(usedVariable, counter)
                    true
                 } else {
                   false
                 }
              }
           } 
         }
      }
      // NEXT X (where X is a variable which becomes a DoubleExpression)
      def NEXT(d:VariableDoubleExpression) = addToMemory {
         memory.next(d.fromVariable)
      }
      // NEXT (without any variable)
      def NEXT = addToMemory {
         memory.next(null)
      }
      
   }

   class LoopRangeExpression(val lower:DoubleExpression) {
      var upper:DoubleExpression = null 
      var step:DoubleExpression = new LiteralDoubleExpression(1.0)
      def TO(u:DoubleExpression) = {
         upper = u
         this
      }
      def STEP(s:DoubleExpression) = {
         step = s
         this
      }
   }
   
   implicit def Int2NewStatement(lineNumber:Int) = new NewStatement(lineNumber)

   implicit def String2StringExpression(s:String) = new LiteralStringExpression(s) 
   implicit def Int2StringExpression(i:Int) = new LiteralStringExpression(if (i>0) ' ' + i.toString else i.toString) 
   
   implicit def Int2JumpTarget(lineNumber:Int) = new LiteralJumpTarget(lineNumber) 
   
   implicit def Double2DoubleExpression(value:Double) = new LiteralDoubleExpression(value) 
   implicit def Int2DoubleExpression(value:Int) = new LiteralDoubleExpression(value.toDouble) 
   
   implicit def DoubleExpression2LoopRangeExpression(value:DoubleExpression) = new LoopRangeExpression(value)
   implicit def Int2LoopRangeExpression(value:Int) = new LoopRangeExpression(new LiteralDoubleExpression(value.toDouble))

   implicit def DoubleExpression2StringExpression(d:DoubleExpression) = 
      new StringExpression {
         def value = if (d.value>0) ' '+d.value.toString else d.value.toString
         override def fromVariable = d.fromVariable
      }

}
